"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.TypeMapper = void 0;

var _parser = require("graphql/language/parser");

var _language = require("graphql/language");

var _buildASTSchema = require("graphql/utilities/buildASTSchema");

var _keyValMap = _interopRequireDefault(require("graphql/jsutils/keyValMap"));

var _invariant = _interopRequireDefault(require("graphql/jsutils/invariant"));

var _values = require("graphql/execution/values");

var _misc = require("./utils/misc");

var _find = _interopRequireDefault(require("./utils/polyfills/find"));

var _graphql = require("./graphql");

var _type = require("./type");

var _createThunkedObjectProxy = require("./utils/createThunkedObjectProxy");

var _ObjectTypeComposer = require("./ObjectTypeComposer");

var _InputTypeComposer = require("./InputTypeComposer");

var _ScalarTypeComposer = require("./ScalarTypeComposer");

var _EnumTypeComposer = require("./EnumTypeComposer");

var _InterfaceTypeComposer = require("./InterfaceTypeComposer");

var _UnionTypeComposer = require("./UnionTypeComposer");

var _ListComposer = require("./ListComposer");

var _NonNullComposer = require("./NonNullComposer");

var _ThunkComposer = require("./ThunkComposer");

var _Resolver = require("./Resolver");

var _TypeStorage = require("./TypeStorage");

var _is = require("./utils/is");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; var ownKeys = Object.keys(source); if (typeof Object.getOwnPropertySymbols === 'function') { ownKeys = ownKeys.concat(Object.getOwnPropertySymbols(source).filter(function (sym) { return Object.getOwnPropertyDescriptor(source, sym).enumerable; })); } ownKeys.forEach(function (key) { _defineProperty(target, key, source[key]); }); } return target; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function _objectWithoutProperties(source, excluded) { if (source == null) return {}; var target = _objectWithoutPropertiesLoose(source, excluded); var key, i; if (Object.getOwnPropertySymbols) { var sourceSymbolKeys = Object.getOwnPropertySymbols(source); for (i = 0; i < sourceSymbolKeys.length; i++) { key = sourceSymbolKeys[i]; if (excluded.indexOf(key) >= 0) continue; if (!Object.prototype.propertyIsEnumerable.call(source, key)) continue; target[key] = source[key]; } } return target; }

function _objectWithoutPropertiesLoose(source, excluded) { if (source == null) return {}; var target = {}; var sourceKeys = Object.keys(source); var key, i; for (i = 0; i < sourceKeys.length; i++) { key = sourceKeys[i]; if (excluded.indexOf(key) >= 0) continue; target[key] = source[key]; } return target; }

class TypeMapper {
  constructor(schemaComposer) {
    if (!schemaComposer) {
      throw new Error('TypeMapper must have SchemaComposer instance.');
    }

    this.schemaComposer = schemaComposer;

    this._initScalars(); // alive proper Flow type casting in autosuggestions for class with Generics

    /* :: return this; */

  }

  _initScalars() {
    this.schemaComposer.add(_graphql.GraphQLString);
    this.schemaComposer.add(_graphql.GraphQLFloat);
    this.schemaComposer.add(_graphql.GraphQLInt);
    this.schemaComposer.add(_graphql.GraphQLBoolean);
    this.schemaComposer.add(_graphql.GraphQLID);
    this.schemaComposer.add(_type.GraphQLJSON);
    this.schemaComposer.add(_type.GraphQLJSONObject);
    this.schemaComposer.add(_type.GraphQLDate);
    this.schemaComposer.add(_type.GraphQLBuffer);
  }

  static isOutputType(type) {
    return type instanceof _ScalarTypeComposer.ScalarTypeComposer || type instanceof _ObjectTypeComposer.ObjectTypeComposer || type instanceof _InterfaceTypeComposer.InterfaceTypeComposer || type instanceof _UnionTypeComposer.UnionTypeComposer || type instanceof _EnumTypeComposer.EnumTypeComposer || type instanceof _NonNullComposer.NonNullComposer && TypeMapper.isOutputType(type.ofType) || type instanceof _ListComposer.ListComposer && TypeMapper.isOutputType(type.ofType) || type instanceof _ThunkComposer.ThunkComposer;
  }

  static isInputType(type) {
    return type instanceof _ScalarTypeComposer.ScalarTypeComposer || type instanceof _EnumTypeComposer.EnumTypeComposer || type instanceof _InputTypeComposer.InputTypeComposer || type instanceof _NonNullComposer.NonNullComposer && TypeMapper.isInputType(type.ofType) || type instanceof _ListComposer.ListComposer && TypeMapper.isInputType(type.ofType) || type instanceof _ThunkComposer.ThunkComposer;
  }

  static isTypeNameString(str) {
    return /^[_A-Za-z][_0-9A-Za-z]*$/.test(str);
  }

  static isTypeDefinitionString(str) {
    return this.isOutputTypeDefinitionString(str) || this.isInputTypeDefinitionString(str) || this.isEnumTypeDefinitionString(str) || this.isScalarTypeDefinitionString(str) || this.isInterfaceTypeDefinitionString(str);
  }

  static isOutputTypeDefinitionString(str) {
    return /type\s[^{]+\{[^}]+\}/im.test(str);
  }

  static isInputTypeDefinitionString(str) {
    return /input\s[^{]+\{[^}]+\}/im.test(str);
  }

  static isEnumTypeDefinitionString(str) {
    return /enum\s[^{]+\{[^}]+\}/im.test(str);
  }

  static isScalarTypeDefinitionString(str) {
    return /scalar\s/im.test(str);
  }

  static isInterfaceTypeDefinitionString(str) {
    return /interface\s/im.test(str);
  }

  convertGraphQLTypeToComposer(type) {
    if (type instanceof _graphql.GraphQLObjectType) {
      return _ObjectTypeComposer.ObjectTypeComposer.create(type, this.schemaComposer);
    } else if (type instanceof _graphql.GraphQLInputObjectType) {
      return _InputTypeComposer.InputTypeComposer.create(type, this.schemaComposer);
    } else if (type instanceof _graphql.GraphQLScalarType) {
      return _ScalarTypeComposer.ScalarTypeComposer.create(type, this.schemaComposer);
    } else if (type instanceof _graphql.GraphQLEnumType) {
      return _EnumTypeComposer.EnumTypeComposer.create(type, this.schemaComposer);
    } else if (type instanceof _graphql.GraphQLInterfaceType) {
      return _InterfaceTypeComposer.InterfaceTypeComposer.create(type, this.schemaComposer);
    } else if (type instanceof _graphql.GraphQLUnionType) {
      return _UnionTypeComposer.UnionTypeComposer.create(type, this.schemaComposer);
    } else if (type instanceof _graphql.GraphQLNonNull) {
      // schema do not store wrapped types
      return new _NonNullComposer.NonNullComposer(this.convertGraphQLTypeToComposer(type.ofType));
    } else if (type instanceof _graphql.GraphQLList) {
      // schema do not store wrapped types
      return new _ListComposer.ListComposer(this.convertGraphQLTypeToComposer(type.ofType));
    }

    throw new Error(`Cannot convert to Composer the following value: ${(0, _misc.inspect)(type)}`);
  }

  convertSDLWrappedTypeName(str) {
    const typeAST = (0, _parser.parseType)(str);
    return this.typeFromAST(typeAST);
  }

  convertSDLTypeDefinition(str) {
    if (this.schemaComposer.has(str)) {
      return this.schemaComposer.getAnyTC(str);
    }

    const astDocument = (0, _parser.parse)(str);

    if (!astDocument || astDocument.kind !== 'Document') {
      throw new Error('You should provide correct type syntax. ' + "Eg. convertSDLTypeDefinition('type IntRange { min: Int, max: Int }')");
    }

    const types = this.parseTypes(astDocument);
    const type = types[0];

    if (type) {
      this.schemaComposer.set(type.getTypeName(), type); // Also keep type string representation for avoiding duplicates type defs for same strings

      this.schemaComposer.set(str, type);
      return type;
    }

    return undefined;
  }

  convertOutputTypeDefinition(typeDef, fieldName = '', typeName = '') {
    if (typeof typeDef === 'string') {
      if (TypeMapper.isInputTypeDefinitionString(typeDef)) {
        throw new Error(`Should be OutputType, but got input type definition: ${(0, _misc.inspect)(typeDef)}'`);
      }

      let tc;

      if (this.schemaComposer.has(typeDef)) {
        tc = this.schemaComposer.getAnyTC(typeDef);
      } else {
        tc = TypeMapper.isTypeDefinitionString(typeDef) ? this.convertSDLTypeDefinition(typeDef) : this.convertSDLWrappedTypeName(typeDef);

        if (!tc) {
          throw new Error(`Cannot convert to OutputType the following string: ${(0, _misc.inspect)(typeDef)}`);
        }
      }

      if (!TypeMapper.isOutputType(tc)) {
        throw new Error(`Provided incorrect OutputType: ${(0, _misc.inspect)(typeDef)}`);
      }

      return tc;
    } else if (TypeMapper.isOutputType(typeDef)) {
      return typeDef;
    } else if (Array.isArray(typeDef)) {
      if (typeDef.length !== 1) {
        throw new Error(`Array must have exact one output type definition, but has ${typeDef.length}: ${(0, _misc.inspect)(typeDef)}`);
      }

      const tc = this.convertOutputTypeDefinition(typeDef[0], fieldName, typeName);

      if (!tc) {
        throw new Error(`Cannot construct TypeComposer from ${(0, _misc.inspect)(typeDef)}`);
      }

      return new _ListComposer.ListComposer(tc);
    } else if ((0, _is.isFunction)(typeDef)) {
      return new _ThunkComposer.ThunkComposer(() => {
        const def = typeDef();
        const tc = this.convertOutputFieldConfig(def, fieldName, typeName).type;

        if (!TypeMapper.isOutputType(tc)) {
          throw new Error(`Provided incorrect OutputType: Function[${(0, _misc.inspect)(def)}]`);
        }

        return tc;
      });
    } else if (typeDef instanceof _Resolver.Resolver) {
      return typeDef.getTypeComposer();
    } else if (typeDef instanceof _graphql.GraphQLList || typeDef instanceof _graphql.GraphQLNonNull) {
      const type = this.convertGraphQLTypeToComposer(typeDef);

      if (TypeMapper.isOutputType(type)) {
        return type;
      } else {
        throw new Error(`Provided incorrect OutputType: ${(0, _misc.inspect)(type)}`);
      }
    } else if (typeDef instanceof _graphql.GraphQLObjectType || typeDef instanceof _graphql.GraphQLEnumType || typeDef instanceof _graphql.GraphQLInterfaceType || typeDef instanceof _graphql.GraphQLUnionType || typeDef instanceof _graphql.GraphQLScalarType) {
      return this.convertGraphQLTypeToComposer(typeDef);
    }

    if (typeDef instanceof _InputTypeComposer.InputTypeComposer) {
      throw new Error(`Should be OutputType, but provided InputTypeComposer ${(0, _misc.inspect)(typeDef)}`);
    }

    return undefined;
  }

  convertOutputFieldConfig(composeFC, fieldName = '', typeName = '') {
    try {
      if (!composeFC) {
        throw new Error(`You provide empty output field definition: ${(0, _misc.inspect)(composeFC)}`);
      }

      if (composeFC instanceof _Resolver.Resolver) {
        return {
          type: composeFC.type,
          args: composeFC.getArgs(),
          resolve: composeFC.getFieldResolver(),
          description: composeFC.getDescription()
        };
      } // use proxy for evaluation on demand


      if ((0, _is.isFunction)(composeFC)) {
        return (0, _createThunkedObjectProxy.createThunkedObjectProxy)(() => this.convertOutputFieldConfig(composeFC(), fieldName, typeName));
      } // convert type when its provided as composeIFC


      const tcFromIFC = this.convertOutputTypeDefinition(composeFC, fieldName, typeName);

      if (tcFromIFC) {
        return {
          type: tcFromIFC
        };
      } // convert type when its provided in composeIFC.type


      if ((0, _is.isObject)(composeFC)) {
        const _ref = composeFC,
              {
          type,
          args
        } = _ref,
              rest = _objectWithoutProperties(_ref, ["type", "args"]);

        if (!type) {
          throw new Error(`Definition object should contain 'type' property: ${(0, _misc.inspect)(composeFC)}`);
        }

        const tc = this.convertOutputTypeDefinition(type, fieldName, typeName);

        if (tc) {
          return _objectSpread({
            type: tc,
            args: this.convertArgConfigMap(args || {}, fieldName, typeName)
          }, rest);
        }
      }

      throw new Error(`Cannot convert to InputType the following value: ${(0, _misc.inspect)(composeFC)}`);
    } catch (e) {
      e.message = `TypeError[${typeName}.${fieldName}]: ${e.message}`;
      throw e;
    }
  }

  convertOutputFieldConfigMap(composeFields, typeName = '') {
    const fields = {};
    Object.keys(composeFields).forEach(name => {
      fields[name] = this.convertOutputFieldConfig(composeFields[name], name, typeName);
    });
    return fields;
  }

  convertArgConfig(composeAC, argName = '', fieldName = '', typeName = '') {
    try {
      if (!composeAC) {
        throw new Error(`You provide empty argument config ${(0, _misc.inspect)(composeAC)}`);
      } // use proxy for evaluation on demand


      if ((0, _is.isFunction)(composeAC)) {
        return (0, _createThunkedObjectProxy.createThunkedObjectProxy)(() => this.convertArgConfig(composeAC(), argName, fieldName, typeName));
      } // convert type when its provided as composeAC


      const tcFromAC = this.convertInputTypeDefinition(composeAC);

      if (tcFromAC) {
        return {
          type: tcFromAC
        };
      } // convert type when its provided in composeIFC.type


      if ((0, _is.isObject)(composeAC)) {
        const _ref2 = composeAC,
              {
          type
        } = _ref2,
              rest = _objectWithoutProperties(_ref2, ["type"]);

        if (!type) {
          throw new Error(`Definition object should contain 'type' property: ${(0, _misc.inspect)(composeAC)}'`);
        }

        const tc = this.convertInputTypeDefinition(type);

        if (tc) {
          return _objectSpread({
            type: tc
          }, rest);
        }
      }

      throw new Error(`Cannot convert to InputType the following value: ${(0, _misc.inspect)(tcFromAC)}`);
    } catch (e) {
      e.message = `TypeError[${typeName}.${fieldName}.${argName}]: ${e.message}`;
      throw e;
    }
  }

  convertArgConfigMap(composeArgsConfigMap, fieldName = '', typeName = '') {
    const argsConfigMap = {};

    if (composeArgsConfigMap) {
      Object.keys(composeArgsConfigMap).forEach(argName => {
        argsConfigMap[argName] = this.convertArgConfig(composeArgsConfigMap[argName], argName, fieldName, typeName);
      });
    }

    return argsConfigMap;
  }

  convertInputTypeDefinition(typeDef, fieldName = '', typeName = '') {
    if (typeof typeDef === 'string') {
      if (TypeMapper.isOutputTypeDefinitionString(typeDef)) {
        throw new Error(`Should be InputType, but got output type definition: ${(0, _misc.inspect)(typeDef)}`);
      }

      let tc;

      if (this.schemaComposer.has(typeDef)) {
        tc = this.schemaComposer.getAnyTC(typeDef);
      } else {
        tc = TypeMapper.isTypeDefinitionString(typeDef) ? this.convertSDLTypeDefinition(typeDef) : this.convertSDLWrappedTypeName(typeDef);

        if (!tc) {
          throw new Error(`Cannot convert to InputType the following string: ${(0, _misc.inspect)(typeDef)}`);
        }
      }

      if (!TypeMapper.isInputType(tc)) {
        throw new Error(`Provided incorrect InputType: ${(0, _misc.inspect)(typeDef)}`);
      }

      return tc;
    } else if (TypeMapper.isInputType(typeDef)) {
      return typeDef;
    } else if (Array.isArray(typeDef)) {
      if (typeDef.length !== 1) {
        throw new Error(`Array must have exact one input type definition, but has ${typeDef.length}: ${(0, _misc.inspect)(typeDef)}`);
      }

      const tc = this.convertInputTypeDefinition(typeDef[0], fieldName, typeName);

      if (!tc) {
        throw new Error(`Cannot construct TypeComposer from ${(0, _misc.inspect)(typeDef)}`);
      }

      return new _ListComposer.ListComposer(tc);
    } else if ((0, _is.isFunction)(typeDef)) {
      return new _ThunkComposer.ThunkComposer(() => {
        const def = typeDef();
        const tc = this.convertInputFieldConfig(def, fieldName, typeName).type;

        if (!TypeMapper.isInputType(tc)) {
          throw new Error(`Provided incorrect InputType: Function[${(0, _misc.inspect)(def)}]`);
        }

        return tc;
      });
    } else if (typeDef instanceof _graphql.GraphQLList || typeDef instanceof _graphql.GraphQLNonNull) {
      const type = this.convertGraphQLTypeToComposer(typeDef);

      if (TypeMapper.isInputType(type)) {
        return type;
      } else {
        throw new Error(`Provided incorrect InputType: ${(0, _misc.inspect)(type)}`);
      }
    } else if (typeDef instanceof _graphql.GraphQLInputObjectType || typeDef instanceof _graphql.GraphQLScalarType || typeDef instanceof _graphql.GraphQLEnumType) {
      return this.convertGraphQLTypeToComposer(typeDef);
    }

    if (typeDef instanceof _ObjectTypeComposer.ObjectTypeComposer) {
      throw new Error(`Should be InputType, but provided ObjectTypeComposer ${(0, _misc.inspect)(typeDef)}`);
    }

    return undefined;
  }

  convertInputFieldConfig(composeIFC, fieldName = '', typeName = '') {
    try {
      if (!composeIFC) {
        throw new Error(`You provide empty input field definition: ${(0, _misc.inspect)(composeIFC)}`);
      } // use proxy for evaluation on demand


      if ((0, _is.isFunction)(composeIFC)) {
        return (0, _createThunkedObjectProxy.createThunkedObjectProxy)(() => this.convertInputFieldConfig(composeIFC(), fieldName, typeName));
      } // convert type when its provided as composeIFC


      const tcFromIFC = this.convertInputTypeDefinition(composeIFC, fieldName, typeName);

      if (tcFromIFC) {
        return {
          type: tcFromIFC
        };
      } // convert type when its provided in composeIFC.type


      if ((0, _is.isObject)(composeIFC)) {
        const _ref3 = composeIFC,
              {
          type
        } = _ref3,
              rest = _objectWithoutProperties(_ref3, ["type"]);

        if (!type) {
          throw new Error(`Definition object should contain 'type' property: ${(0, _misc.inspect)(composeIFC)}`);
        }

        const tc = this.convertInputTypeDefinition(type, fieldName, typeName);

        if (tc) {
          return _objectSpread({
            type: tc
          }, rest);
        }
      }

      throw new Error(`Cannot convert to InputType the following value: ${(0, _misc.inspect)(composeIFC)}`);
    } catch (e) {
      e.message = `TypeError[${typeName}.${fieldName}]: ${e.message}`;
      throw e;
    }
  }

  convertInputFieldConfigMap(composeFields, typeName = '') {
    const fields = {};
    Object.keys(composeFields).forEach(name => {
      fields[name] = this.convertInputFieldConfig(composeFields[name], name, typeName);
    });
    return fields;
  }

  convertInterfaceTypeDefinition(typeDef) {
    if (this.schemaComposer.hasInstance(typeDef, _InterfaceTypeComposer.InterfaceTypeComposer)) {
      return this.schemaComposer.getIFTC(typeDef);
    } else if (typeof typeDef === 'string') {
      const tc = TypeMapper.isInterfaceTypeDefinitionString(typeDef) ? this.convertSDLTypeDefinition(typeDef) : this.convertSDLWrappedTypeName(typeDef);

      if (!(tc instanceof _InterfaceTypeComposer.InterfaceTypeComposer) && !(tc instanceof _ThunkComposer.ThunkComposer)) {
        throw new Error(`Cannot convert to InterfaceType the following definition: ${(0, _misc.inspect)(typeDef)}`);
      }

      return tc;
    } else if (typeDef instanceof _graphql.GraphQLInterfaceType) {
      return new _InterfaceTypeComposer.InterfaceTypeComposer(typeDef, this.schemaComposer);
    } else if (typeDef instanceof _InterfaceTypeComposer.InterfaceTypeComposer || typeDef instanceof _ThunkComposer.ThunkComposer) {
      return typeDef;
    } else if ((0, _is.isFunction)(typeDef)) {
      return new _ThunkComposer.ThunkComposer(() => this.convertInterfaceTypeDefinition(typeDef()));
    }

    throw new Error(`Cannot convert to InterfaceType the following definition: ${(0, _misc.inspect)(typeDef)}`);
  }

  parseTypesFromString(str) {
    const astDocument = (0, _parser.parse)(str);

    if (!astDocument || astDocument.kind !== 'Document') {
      throw new Error('You should provide correct SDL syntax.');
    }

    const types = this.parseTypes(astDocument);
    const typeStorage = new _TypeStorage.TypeStorage();
    types.forEach(type => {
      typeStorage.set(type.getTypeName(), type);
    });
    return typeStorage;
  }
  /**
   * -----------------------------------------------
   * Internal methods
   * -----------------------------------------------
   */


  parseTypes(astDocument) {
    const types = [];

    for (let i = 0; i < astDocument.definitions.length; i++) {
      const def = astDocument.definitions[i];
      const type = this.makeSchemaDef(def);

      if (type) {
        types[i] = type;
      }
    }

    return types;
  }

  typeFromAST(typeNode) {
    let innerType;

    if (typeNode.kind === _language.Kind.LIST_TYPE) {
      return new _ListComposer.ListComposer(this.typeFromAST(typeNode.type));
    } else if (typeNode.kind === _language.Kind.NON_NULL_TYPE) {
      return new _NonNullComposer.NonNullComposer(this.typeFromAST(typeNode.type));
    }

    (0, _invariant.default)(typeNode.kind === _language.Kind.NAMED_TYPE, `Must be a named type for ${(0, _misc.inspect)(typeNode)}.`);
    const typeName = typeNode.name.value;

    if (this.schemaComposer.has(typeName)) {
      return this.schemaComposer.get(typeName);
    } else {
      return new _ThunkComposer.ThunkComposer(() => {
        return this.schemaComposer.get(typeName);
      }, typeName);
    }
  }

  typeFromASTInput(typeNode) {
    const tc = this.typeFromAST(typeNode);

    if (!TypeMapper.isInputType(tc)) {
      throw new Error(`TypeAST should be for Input types. But recieved ${(0, _misc.inspect)(typeNode)}`);
    }

    return tc;
  }

  typeFromASTOutput(typeNode) {
    const tc = this.typeFromAST(typeNode);

    if (!TypeMapper.isOutputType(tc)) {
      throw new Error(`TypeAST should be for Output types. But recieved ${(0, _misc.inspect)(typeNode)}`);
    }

    return tc;
  }

  makeSchemaDef(def) {
    if (!def) {
      throw new Error('def must be defined');
    }

    switch (def.kind) {
      case _language.Kind.OBJECT_TYPE_DEFINITION:
        return this.makeTypeDef(def);

      case _language.Kind.INTERFACE_TYPE_DEFINITION:
        return this.makeInterfaceDef(def);

      case _language.Kind.ENUM_TYPE_DEFINITION:
        return this.makeEnumDef(def);

      case _language.Kind.UNION_TYPE_DEFINITION:
        return this.makeUnionDef(def);

      case _language.Kind.SCALAR_TYPE_DEFINITION:
        return this.makeScalarDef(def);

      case _language.Kind.SCHEMA_DEFINITION:
        this.checkSchemaDef(def);
        return null;

      case _language.Kind.DIRECTIVE_DEFINITION:
        {
          const directive = this.makeDirectiveDef(def);
          if (directive) this.schemaComposer.addDirective(directive);
          return null;
        }

      case _language.Kind.INPUT_OBJECT_TYPE_DEFINITION:
        return this.makeInputObjectDef(def);

      default:
        throw new Error(`Type kind "${def.kind}" not supported.`);
    }
  }

  makeArguments(values) {
    if (!values) {
      return {};
    }

    const result = {};
    values.forEach(value => {
      const key = value.name.value;
      let val;
      const typeName = this.getNamedTypeAST(value.type).name.value;
      const type = this.typeFromASTInput(value.type);
      const ac = {
        type,
        description: (0, _buildASTSchema.getDescription)(value)
      };

      if (value.directives) {
        const directives = this.parseDirectives(value.directives);

        if (directives) {
          ac.extensions = {
            directives
          };
          const data = directives.find(d => d.name === 'default');

          if (data) {
            ac.defaultValue = data.args.value;
          }
        }
      }

      if (value.defaultValue) {
        const typeDef = this.schemaComposer.get(typeName);
        const wrappedType = this.buildWrappedTypeDef(typeDef, value.type);

        if (TypeMapper.isInputType(wrappedType)) {
          ac.defaultValue = (0, _graphql.valueFromAST)(value.defaultValue, wrappedType.getType());
        } else {
          throw new Error('Non-input type as an argument.');
        }
      }

      result[key] = ac;
    });
    return result;
  }

  makeFieldDefMap(def) {
    if (!def.fields) return {};
    return (0, _keyValMap.default)(def.fields, field => field.name.value, field => {
      const fc = {
        type: this.typeFromASTOutput(field.type),
        description: (0, _buildASTSchema.getDescription)(field),
        args: this.makeArguments(field.arguments),
        deprecationReason: this.getDeprecationReason(field.directives),
        astNode: field
      };

      if (field.directives) {
        const directives = this.parseDirectives(field.directives);

        if (directives) {
          fc.extensions = {
            directives
          };
        }
      }

      return fc;
    });
  }

  makeInputFieldDef(def) {
    if (!def.fields) return {};
    return (0, _keyValMap.default)(def.fields, field => field.name.value, field => {
      const fc = {
        type: this.typeFromASTInput(field.type),
        description: (0, _buildASTSchema.getDescription)(field),
        deprecationReason: this.getDeprecationReason(field.directives),
        astNode: field
      };

      if (field.directives) {
        const directives = this.parseDirectives(field.directives);

        if (directives) {
          fc.extensions = {
            directives
          };
          const data = directives.find(d => d.name === 'default');

          if (data) {
            fc.defaultValue = data.args.value;
          }
        }
      }

      return fc;
    });
  }

  makeEnumDef(def) {
    const tc = this.schemaComposer.createEnumTC({
      name: def.name.value,
      description: (0, _buildASTSchema.getDescription)(def),
      values: !def.values ? {} : (0, _keyValMap.default)(def.values, enumValue => enumValue.name.value, enumValue => {
        const ec = {
          description: (0, _buildASTSchema.getDescription)(enumValue),
          deprecationReason: this.getDeprecationReason(enumValue.directives)
        };

        if (enumValue.directives) {
          const directives = this.parseDirectives(enumValue.directives);

          if (directives) {
            ec.extensions = {
              directives
            };
          }
        }

        return ec;
      }),
      astNode: def
    });

    if (def.directives) {
      tc.setExtension('directives', this.parseDirectives(def.directives));
    }

    return tc;
  }

  makeInputObjectDef(def) {
    const tc = this.schemaComposer.createInputTC({
      name: def.name.value,
      description: (0, _buildASTSchema.getDescription)(def),
      fields: this.makeInputFieldDef(def),
      astNode: def
    });

    if (def.directives) {
      tc.setExtension('directives', this.parseDirectives(def.directives));
    }

    return tc;
  }

  makeDirectiveDef(def) {
    const locations = def.locations.map(({
      value
    }) => value);
    const args = {};
    (def.arguments || []).forEach(value => {
      const key = value.name.value;
      let val;
      const wrappedType = this.typeFromAST(value.type);

      if (TypeMapper.isInputType(wrappedType)) {
        val = {
          type: wrappedType,
          description: (0, _buildASTSchema.getDescription)(value)
        };
      } else {
        throw new Error('Non-input type as an argument.');
      }

      args[key] = val;
    });
    return new _graphql.GraphQLDirective({
      name: def.name.value,
      description: (0, _buildASTSchema.getDescription)(def),
      locations,
      args,
      astNode: def
    });
  }

  makeScalarDef(def) {
    const tc = this.schemaComposer.createScalarTC({
      name: def.name.value,
      description: (0, _buildASTSchema.getDescription)(def),
      serialize: v => v,
      astNode: def
    });

    if (def.directives) {
      tc.setExtension('directives', this.parseDirectives(def.directives));
    }

    return tc;
  }

  makeImplementedInterfaces(def) {
    return (def.interfaces || []).map(iface => {
      const name = this.getNamedTypeAST(iface).name.value;

      if (this.schemaComposer.hasInstance(name, _InterfaceTypeComposer.InterfaceTypeComposer)) {
        return this.schemaComposer.getIFTC(name);
      } else {
        return new _ThunkComposer.ThunkComposer(() => this.schemaComposer.getIFTC(name), name);
      }
    });
  }

  makeTypeDef(def) {
    const tc = this.schemaComposer.createObjectTC({
      name: def.name.value,
      description: (0, _buildASTSchema.getDescription)(def),
      fields: this.makeFieldDefMap(def),
      interfaces: this.makeImplementedInterfaces(def),
      astNode: def
    });

    if (def.directives) {
      tc.setExtension('directives', this.parseDirectives(def.directives));
    }

    return tc;
  }

  makeInterfaceDef(def) {
    const tc = this.schemaComposer.createInterfaceTC({
      name: def.name.value,
      description: (0, _buildASTSchema.getDescription)(def),
      fields: this.makeFieldDefMap(def),
      astNode: def
    });

    if (def.directives) {
      tc.setExtension('directives', this.parseDirectives(def.directives));
    }

    return tc;
  }

  makeUnionDef(def) {
    const types = def.types;
    const tc = this.schemaComposer.createUnionTC({
      name: def.name.value,
      description: (0, _buildASTSchema.getDescription)(def),
      types: (types || []).map(ref => this.getNamedTypeAST(ref).name.value),
      astNode: def
    });

    if (def.directives) {
      tc.setExtension('directives', this.parseDirectives(def.directives));
    }

    return tc;
  }

  checkSchemaDef(def) {
    const validNames = {
      query: 'Query',
      mutation: 'Mutation',
      subscription: 'Subscription'
    };
    def.operationTypes.forEach(d => {
      if (d.operation) {
        const validTypeName = validNames[d.operation];
        const actualTypeName = d.type.name.value;

        if (actualTypeName !== validTypeName) {
          throw new Error(`Incorrect type name '${actualTypeName}' for '${d.operation}'. The valid definition is "schema { ${d.operation}: ${validTypeName} }"`);
        }
      }
    });
  }

  getNamedTypeAST(typeAST) {
    let namedType = typeAST;

    while (namedType.kind === _language.Kind.LIST_TYPE || namedType.kind === _language.Kind.NON_NULL_TYPE) {
      namedType = namedType.type;
    }

    return namedType;
  }

  buildWrappedTypeDef(innerType, inputTypeAST) {
    if (inputTypeAST.kind === _language.Kind.LIST_TYPE) {
      return new _ListComposer.ListComposer(this.buildWrappedTypeDef(innerType, inputTypeAST.type));
    }

    if (inputTypeAST.kind === _language.Kind.NON_NULL_TYPE) {
      const wrappedType = this.buildWrappedTypeDef(innerType, inputTypeAST.type);
      return new _NonNullComposer.NonNullComposer(wrappedType);
    }

    return innerType;
  }

  getDeprecationReason(directives) {
    const deprecatedAST = directives && (0, _find.default)(directives, directive => directive.name.value === _graphql.GraphQLDeprecatedDirective.name);

    if (!deprecatedAST) {
      return;
    }

    const {
      reason
    } = (0, _values.getArgumentValues)(_graphql.GraphQLDeprecatedDirective, deprecatedAST);
    return reason; // eslint-disable-line
  }

  parseDirectives(directives) {
    const result = [];
    directives.forEach(directive => {
      const name = directive.name.value;

      const directiveDef = this.schemaComposer._getDirective(name);

      const args = directiveDef ? (0, _values.getArgumentValues)(directiveDef, directive) : (0, _keyValMap.default)(directive.arguments || [], arg => arg.name.value, arg => _type.GraphQLJSON.parseLiteral(arg.value));
      result.push({
        name,
        args
      });
    });
    return result;
  }

}

exports.TypeMapper = TypeMapper;